"""Application programming interface for the jazzy package."""
# src/jazzy/api.py
import base64

from jazzy.config import Config
from jazzy.config import MINIMISATION_METHOD
from jazzy.config import ROUNDING_DIGITS
from jazzy.core import calculate_delta_apolar
from jazzy.core import calculate_delta_interaction
from jazzy.core import calculate_delta_polar
from jazzy.core import calculate_polar_strength_map
from jazzy.core import get_charges_from_kallisto_molecule
from jazzy.core import get_covalent_atom_idxs
from jazzy.core import kallisto_molecule_from_rdkit_molecule
from jazzy.core import rdkit_molecule_from_smiles
from jazzy.helpers import condense_atomic_map
from jazzy.helpers import convert_map_to_tuples
from jazzy.helpers import sum_atomic_map
from jazzy.utils import JazzyError
from jazzy.visualisation import depict_strengths


# global jazzy config (parameter)
config = Config()


def __smiles_to_molecule_objects(smiles, minimisation_method=MINIMISATION_METHOD):
    """Private method for converting SMILES into RDKit and kallisto objects."""
    rdkit_mol = rdkit_molecule_from_smiles(
        smiles, minimisation_method=minimisation_method
    )
    if rdkit_mol is None:
        raise JazzyError("The SMILES '{}' could not be processed.".format(smiles))
    kallisto_mol = kallisto_molecule_from_rdkit_molecule(rdkit_mol)
    return rdkit_mol, kallisto_mol


def molecular_vector_from_smiles(
    smiles: str, minimisation_method=MINIMISATION_METHOD, only_strengths=False
):
    """API route to calculate molecular free energy vector.

    Calculates the apolar (dga), the polar (dgp), and the interaction (dgi)
    contribution to the free energy.

    Args:
    smiles: A molecule SMILES string representation (default '')
    minimisation_method: One of the conformer energy minimisation methods
    as available in RDKit (available is 'MMFF94', 'MMFF94s', or 'UFF') (default None)
    only_strengths: Boolean value that determines wheather to calculate only
    strengts or even more.

    Returns:
    Molecular strength vector with or without free energy contributions.

    """
    mol_vector = dict()
    # generate an RDKit molecule
    rdkit_molecule, kallisto_molecule = __smiles_to_molecule_objects(
        smiles, minimisation_method
    )
    atoms_and_nbrs = get_covalent_atom_idxs(rdkit_molecule)
    kallisto_charges = get_charges_from_kallisto_molecule(kallisto_molecule, 0)
    atomic_map = calculate_polar_strength_map(
        rdkit_molecule, kallisto_molecule, atoms_and_nbrs, kallisto_charges
    )
    mol_vector = sum_atomic_map(atomic_map)

    # add free energy contributions
    if not only_strengths:
        dg = dict()
        dg["dga"] = calculate_delta_apolar(
            rdkit_molecule,
            atomic_map,
            config.g0,
            config.gs,
            config.gr,
            config.gpi1,
            config.gpi2,
        )
        dg["dgp"] = calculate_delta_polar(
            atomic_map, atoms_and_nbrs, config.gd, config.ga, config.expd, config.expa
        )
        dgi = calculate_delta_interaction(
            rdkit_molecule, atomic_map, atoms_and_nbrs, config.gi, config.expa, config.f
        )
        dg["dgtot"] = sum(dg.values()) + dgi
        dg = {k: round(dg[k], ROUNDING_DIGITS) for k in dg}
        mol_vector = {**mol_vector, **dg}  # type: ignore
    return mol_vector


def deltag_from_smiles(smiles: str, minimisation_method=MINIMISATION_METHOD):
    """API route to calculate molecular free energy scalar.

    Args:
    smiles: A molecule SMILES string representation (default '')
    minimisation_method: One of the conformer energy minimisation methods
    as available in RDKit (available is 'MMFF94', 'MMFF94s', or 'UFF') (default None)

    Returns:
    Free energy as scalar rounded.

    """
    # generate basic descriptors
    rdkit_molecule, kallisto_molecule = __smiles_to_molecule_objects(
        smiles, minimisation_method
    )
    atoms_and_nbrs = get_covalent_atom_idxs(rdkit_molecule)
    kallisto_charges = get_charges_from_kallisto_molecule(kallisto_molecule, 0)
    atomic_map = calculate_polar_strength_map(
        rdkit_molecule, kallisto_molecule, atoms_and_nbrs, kallisto_charges
    )

    # generate free energy scalar
    dg = dict()
    dg["dga"] = calculate_delta_apolar(
        rdkit_molecule,
        atomic_map,
        config.g0,
        config.gs,
        config.gr,
        config.gpi1,
        config.gpi2,
    )
    dg["dgp"] = calculate_delta_polar(
        atomic_map, atoms_and_nbrs, config.gd, config.ga, config.expd, config.expa
    )
    dg["dgi"] = calculate_delta_interaction(
        rdkit_molecule, atomic_map, atoms_and_nbrs, config.gi, config.expa, config.f
    )
    return round(sum(dg.values()), ROUNDING_DIGITS)


def atomic_tuples_from_smiles(smiles: str, minimisation_method=MINIMISATION_METHOD):
    """API route to generate a tuple representation on the atomic map.

    Not recommended if serialization is needed.

    Args:
    smiles: A molecule SMILES string representation (default '')
    minimisation_method: One of the conformer energy minimisation methods
    as available in RDKit (available is 'MMFF94', 'MMFF94s', or 'UFF') (default None)

    Returns:
    Tuple representation of the atomic map.

    """
    # generate basic descriptors
    rdkit_molecule, kallisto_molecule = __smiles_to_molecule_objects(
        smiles, minimisation_method
    )
    atoms_and_nbrs = get_covalent_atom_idxs(rdkit_molecule)
    kallisto_charges = get_charges_from_kallisto_molecule(kallisto_molecule, 0)
    atomic_map = calculate_polar_strength_map(
        rdkit_molecule, kallisto_molecule, atoms_and_nbrs, kallisto_charges
    )
    return convert_map_to_tuples(atomic_map)


def atomic_map_from_smiles(smiles: str, minimisation_method=MINIMISATION_METHOD):
    """API route to generate a condensed representation on the atomic map.

    Recommended if serialization is needed.

    Args:
    smiles: A molecule SMILES string representation (default '')
    minimisation_method: One of the conformer energy minimisation methods
    as available in RDKit (available is 'MMFF94', 'MMFF94s', or 'UFF') (default None)

    Returns:
    Condensed representation of the atomic map.

    """
    # generate basic descriptors
    rdkit_molecule, kallisto_molecule = __smiles_to_molecule_objects(
        smiles, minimisation_method
    )
    atoms_and_nbrs = get_covalent_atom_idxs(rdkit_molecule)
    kallisto_charges = get_charges_from_kallisto_molecule(kallisto_molecule, 0)
    atomic_map = calculate_polar_strength_map(
        rdkit_molecule, kallisto_molecule, atoms_and_nbrs, kallisto_charges
    )
    return condense_atomic_map(atomic_map)


def atomic_strength_vis_from_smiles(
    smiles: str,
    minimisation_method=MINIMISATION_METHOD,
    encode=False,
    fig_size=(500, 500),
    flatten_molecule=False,
    highlight_atoms=False,
    ignore_sdc=False,
    ignore_sdx=False,
    ignore_sa=False,
    sdc_threshold=0.0,
    sdx_threshold=0.0,
    sa_threshold=0.0,
    rounding_digits=ROUNDING_DIGITS,
):
    """API route to generate an SVG image from SMILES string.

    Args:
    smiles: A molecule SMILES string representation (default '')
    minimisation_method: One of the conformer energy minimisation methods
    as available in RDKit (available is 'MMFF94', 'MMFF94s', or 'UFF') (default None)

    Returns:
    SVG image either 2D or 3D.

    """
    # generate basic descriptors
    rdkit_molecule, kallisto_molecule = __smiles_to_molecule_objects(
        smiles, minimisation_method
    )
    atoms_and_nbrs = get_covalent_atom_idxs(rdkit_molecule)
    kallisto_charges = get_charges_from_kallisto_molecule(kallisto_molecule, 0)
    atomic_map = calculate_polar_strength_map(
        rdkit_molecule, kallisto_molecule, atoms_and_nbrs, kallisto_charges
    )
    img_txt = depict_strengths(
        rdkit_molecule=rdkit_molecule,
        atomic_map=atomic_map,
        fig_size=fig_size,
        flatten_molecule=flatten_molecule,
        highlight_atoms=highlight_atoms,
        ignore_sdc=ignore_sdc,
        ignore_sdx=ignore_sdx,
        ignore_sa=ignore_sa,
        sdc_threshold=sdc_threshold,
        sdx_threshold=sdx_threshold,
        sa_threshold=sa_threshold,
        rounding_digits=rounding_digits,
    )
    if encode:
        img_txt = base64.b64encode(img_txt.encode("utf-8"))
    return img_txt



"""Data structure helpers. Anything that is not chemoinformatics goes here."""
# src/jazzy/helpers.py


def condense_atomic_map(atomic_map: dict) -> list:
    """Create condensed representation of the polar strength map.

    Polar strength map generated from `calculate_polar_strength_map()`.

    Args:
    atomic_map: polar strength map for all atoms in the system.

    """
    condensed_map = list()
    for idx, props in atomic_map.items():
        props["idx"] = idx
        condensed_map.append(props)
    return condensed_map


def convert_map_to_tuples(atomic_map: dict) -> list:
    """Create tuple representation of polar strength map.

    Tuple representation of the polar strength map generated from
    `calculate_polar_strength_map()`. Simple example:

    Args:
    atomic_map: polar strength map for all atoms in the system.

    Returns:
    List of tuples, where elements are atom indices and tuples are tuples of
    properties.

    """
    atomic_map_values = atomic_map.values()
    if len(atomic_map_values) == 0:
        raise IndexError("The atomic map must have length greater than zero.")

    tuple_map = list()
    for props in atomic_map_values:
        tuple_map.append(tuple((k, v) for k, v in props.items()))
    return tuple_map


def sum_atomic_map(atomic_map: dict) -> dict:
    """Optimized for summing the values within a polar strength map."""
    atomic_map_values = atomic_map.values()
    if len(atomic_map_values) == 0:
        raise IndexError("The atomic map must have length greater than zero.")

    # sum and round the countable values
    dct = dict()
    countable = ["sdc", "sdx", "sa"]
    for k in countable:
        # multiply for number of lone pairs before summing up for acceptors
        if k == "sa":
            dct[k] = sum(d[k] * d["num_lp"] for d in atomic_map_values if d)
        else:
            dct[k] = sum(d[k] for d in atomic_map_values if d)
    return {key: round(dct[key], 4) for key in dct}



"""Global config file of jazzy package."""
# src/jazzy/config.py

MINIMISATION_METHOD = None
ROUNDING_DIGITS = 4
ANNOTATION_FONT_SCALE = 0.7


class Config(object):
    """Global config object.

    Global jazzy config file for command-line interface (CLI).
    Passable to CLI methods via '@pass_config' decorator.
    Init method initializes the global parameters that are neccessary within
    the calculation of free energy contributions.

    Parameter in config:
    self.g0: apolar zeroth order parameter
    self.gs: apolar surface parameter
    self.gr: apolar ring parameter
    self.gpi1: apolar first pi parameter
    self.gpi2: apolar second pi parameter
    self.gd: hydrogen bond donor parameter
    self.ga: hydrogen bond acceptor parameter
    self.expd: exponent hydrogen bond donor
    self.expa: exponent hydrogen bond acceptor
    self.gi: interaction parameter
    self.smiles: SMILES string of system

    """

    def __init__(
        self,
        g0=1.884,
        gs=0.0467,
        gr=-3.643,
        gpi1=-1.602,
        gpi2=-1.174,
        gd=-0.908,
        ga=-16.131,
        expd=0.50,
        expa=0.34,
        gi=4.9996,
        f=0.514,
    ):
        """Initialize global config."""
        self.g0 = g0
        self.gs = gs
        self.gr = gr
        self.gpi1 = gpi1
        self.gpi2 = gpi2
        self.gd = gd
        self.ga = ga
        self.expd = expd
        self.expa = expa
        self.gi = gi
        self.f = f
        self.smiles = None



"""Main jazzy program."""
from typing import Tuple

import click

from jazzy.api import atomic_strength_vis_from_smiles
from jazzy.api import molecular_vector_from_smiles
from jazzy.config import Config
from jazzy.utils import JazzyError

pass_config = click.make_pass_decorator(Config, ensure=True)


@click.group()
def cli():
    """Command line interface of Jazzy."""


# CLI subcommands


@cli.command("vec")
@pass_config
@click.option(
    "--opt",
    default=None,
    type=str,
    show_default=True,
    help="Optimisation method (None, MMFF94, MMFF94s, or UFF).",
)
@click.option("--strength_only", is_flag=True)
@click.argument("smiles", type=str, default=None, required=True)
def vec(config, smiles: str, opt: str, strength_only: bool) -> None:
    """Calculate molecular descriptors."""
    # raise ValueError if minimisation method is not valid
    valid_methods = [None, "MMFF94", "MMFF94s", "UFF"]
    if opt not in valid_methods:
        raise ValueError(
            f"Minimisation method within '--opt' flag is not valid {valid_methods}"
        )

    # write SMILE to global config
    config.smiles = smiles

    # obtain molecular vector via API
    mol_vector = dict()
    try:
        mol_vector = molecular_vector_from_smiles(
            smiles=smiles,
            minimisation_method=opt,
            only_strengths=strength_only,
        )
        mol_vector["__status"] = "success"
        mol_vector["smiles"] = config.smiles
    except JazzyError:
        mol_vector["__status"] = "error"
        mol_vector["smiles"] = config.smiles
    print(mol_vector)


@cli.command("vis")
@pass_config
@click.option(
    "--opt",
    default=None,
    type=str,
    show_default=True,
    help="Optimisation method (None, MMFF94, MMFF94s, or UFF).",
)
@click.option(
    "--fig_size",
    default=(500, 500),
    type=(int, int),
    show_default=True,
    help="Size of SVG image in pixels.",
)
@click.option(
    "--sdc_threshold",
    default=0.0,
    type=float,
    show_default=True,
    help="Treshold strength to depic Carbon donors.",
)
@click.option(
    "--sdx_threshold",
    default=0.0,
    type=float,
    show_default=True,
    help="Treshold strength to depic non-Carbon donors.",
)
@click.option(
    "--sa_threshold",
    default=0.0,
    type=float,
    show_default=True,
    help="Treshold strength to depic acceptors.",
)
@click.option("--base64", is_flag=True)
@click.option("--flatten_molecule", is_flag=True)
@click.option("--highlight_atoms", is_flag=True)
@click.option("--ignore_sdc", is_flag=True)
@click.option("--ignore_sdx", is_flag=True)
@click.option("--ignore_sa", is_flag=True)
@click.argument("smiles", type=str, default=None, required=True)
def vis(
    config,
    smiles: str,
    opt: str,
    fig_size: Tuple[int, int],
    base64: bool,
    flatten_molecule: bool,
    highlight_atoms: bool,
    ignore_sdc: bool,
    ignore_sdx: bool,
    ignore_sa: bool,
    sdc_threshold: float,
    sdx_threshold: float,
    sa_threshold: float,
):
    """Create SVG image."""
    # raise ValueError if minimisation method is not valid
    valid_methods = [None, "MMFF94", "MMFF94s", "UFF"]
    if opt not in valid_methods:
        raise ValueError(
            f"Minimisation method within '--opt' flag is not valid {valid_methods}"
        )

    # write SMILE to global config
    config.smiles = smiles

    # obtain visualisation vector via API
    vis_vector = dict()
    try:
        svg = atomic_strength_vis_from_smiles(
            smiles=smiles,
            minimisation_method=opt,
            encode=base64,
            fig_size=fig_size,
            flatten_molecule=flatten_molecule,
            highlight_atoms=highlight_atoms,
            ignore_sdc=ignore_sdc,
            ignore_sdx=ignore_sdx,
            ignore_sa=ignore_sa,
            sdc_threshold=sdc_threshold,
            sdx_threshold=sdx_threshold,
            sa_threshold=sa_threshold,
        )
        vis_vector["svg"] = svg
        vis_vector["smiles"] = config.smiles
        vis_vector["__status"] = "success"
    except JazzyError:
        vis_vector["__status"] = "error"
        vis_vector["smiles"] = config.smiles
    print(vis_vector)



"""Core functions of the jazzy package."""
# src/jazzy/core.py
import logging

import numpy as np
from kallisto.atom import Atom
from kallisto.methods import getPolarizabilities
from kallisto.methods import getVanDerWaalsRadii
from kallisto.molecule import Molecule
from kallisto.units import Bohr
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem import GetPeriodicTable
from rdkit.Chem import PeriodicTable
from rdkit.Chem import rdchem
from rdkit.Chem import rdMolDescriptors

from jazzy.config import ROUNDING_DIGITS
from jazzy.utils import KallistoError


logging.basicConfig(
    format="Jazzy %(levelname)s: [%(asctime)s] %(message)s",
    level=logging.WARNING,
    datefmt="%H:%M:%S",
)


def rdkit_molecule_from_smiles(smiles: str, minimisation_method=None):
    """Molecule preparation: Parse SMILES, add hydrogens, and does energy minimisation.

    Args:
    smiles: A molecule SMILES string representation (default '')
    minimisation_method: One of the conformer energy minimisation methods
    as available in RDKit (available is 'MMFF94', 'MMFF94s', or 'UFF') (default None)

    Returns:
    An RDKit molecule (rdkit.Chem.rdchem.Mol) or None if the process fails

    Returns:
    ValueError: If 'minimisation_method' does not correspond to one of the
    available methods.

    """
    # create molecule, add hydrogens, and generate embedding
    m = Chem.MolFromSmiles(smiles)
    if m is None:
        logging.error(
            "The RDKit SMILES parsing has failed for the molecule: %s", smiles
        )
        return None
    mh = Chem.AddHs(m)
    emb_code = AllChem.EmbedMolecule(mh, randomSeed=11)
    if emb_code == -1:
        logging.error("The RDKit embedding has failed for the molecule: %s", smiles)
        return None

    # energy minimisation
    valid_methods = [None, "MMFF94", "MMFF94s", "UFF"]
    if minimisation_method is not None:
        if minimisation_method not in valid_methods:
            raise ValueError(f"Select a valid minimisation method {valid_methods}")
        if minimisation_method == valid_methods[1]:
            AllChem.MMFFOptimizeMolecule(mh)
        if minimisation_method == valid_methods[2]:
            AllChem.MMFFOptimizeMolecule(mh, mmffVariant="MMFF94s")
        if minimisation_method == valid_methods[3]:
            AllChem.UFFOptimizeMolecule(mh)
    return mh


def get_covalent_atom_idxs(rdkit_molecule: Chem.rdchem.Mol) -> list:
    """Get covalent indices for atom_idx.

    Creates a list of lists, where indices are atom indices and the content
    of the lists are indices of binding atoms (e.g., [[1],[0]] means that atom
    0 is bound to atom 1 ([1]) and vice versa ([0]).

    """
    atoms_and_idxs = list()
    for _, atom in enumerate(rdkit_molecule.GetAtoms()):
        nbrs = atom.GetNeighbors()
        atoms_and_idxs.append(list(map(lambda x: x.GetIdx(), nbrs)))
    return atoms_and_idxs


def get_all_neighbors(rdkit_molecule: Chem.rdchem.Mol, molecule_covalent_nbrs: list):
    """Get all alpha, beta, and gamma neighbours.

    Create dictionaries for covalent bonding partner (alpha), all nearest
    neighbours (beta), and nearest-nearest (gamma) neighbours. Every dictionary
    contains the atomic index within the molecule as key and a list of
    neighbours as the value.

    Args:
    rdkit_molecule: RDKit molecule

    Returns:
    Dictionaries for alpha, beta, and gamma neighbours

    """
    alpha = dict()
    beta = dict()
    gamma = dict()
    for i, _ in enumerate(rdkit_molecule.GetAtoms()):
        i = int(i)
        alpha[i], beta[i], gamma[i] = get_atom_neighbours(i, molecule_covalent_nbrs)
    return alpha, beta, gamma


def kallisto_molecule_from_rdkit_molecule(rdkit_molecule: Chem.rdchem.Mol) -> Molecule:
    """Create a kallisto molecule from RDKit molecule.

    Args:
    rdkit_molecule: RDKit molecule

    Returns:
    A kallisto molecule (kallisto.molecule.Molecule)

    Raises:
    KallistoError: An error if the kallisto molecule cannot be created

    """
    # get the name of the molecule if it comes from SDF
    name = ""
    if rdkit_molecule.HasProp("_Name"):
        name = rdkit_molecule.GetProp("_Name")
    # get all xyz coordinates and split into list of lines
    xyz = Chem.rdmolfiles.MolToXYZBlock(rdkit_molecule).split("\n")
    # remove empty lines or molecule name from list
    xyz = [string for string in xyz if string != "" and string != name]
    # remove number of atoms as given in xmol files (first line)
    xyz = xyz[1:]

    # setup periodic table
    pt = GetPeriodicTable()
    # create list of atoms
    atoms = []
    # create kallisto molecule
    for coord in xyz:
        elem, x, y, z = coord.split()[:4]

        # convert atomic coordinates from Angstrom to Bohr
        position = [float(x) / Bohr, float(y) / Bohr, float(z) / Bohr]
        atom = Atom(symbol=pt.GetAtomicNumber(elem), position=position)
        atoms.append(atom)
    kallisto_mol = Molecule(symbols=atoms)
    if "numbers" not in kallisto_mol.arrays.keys():
        raise KallistoError(
            "The kallisto molecule was not created for the input '{}'".format(
                Chem.MolToSmiles(rdkit_molecule)
            )
        )
    return kallisto_mol


def get_charges_from_kallisto_molecule(
    kallisto_molecule: Molecule, charge: int
) -> list:
    """Calculate electronegativity equilibration (EEQ) atomic partial charges.

    Args:
    kallisto_molecule: kallisto molecule
    charge: molecular charge (int)

    Returns:
    List of electronegativity equilibration atomic partial charges

    """
    return list(kallisto_molecule.get_eeq(charge=charge))


def calculate_polar_strength_map(
    rdkit_molecule: Chem.rdchem.Mol,
    kallisto_molecule: Molecule,
    atoms_and_nbrs: list,
    charges: list,
    d=6.1475,
    a=-2.2316,
    t=0.274,
) -> dict:
    """Calculate the polar strength map.

    Generates a molecular dictionary where keys correspond to atom indices,
    and values are dictionaries of calculated properties. These properties
    include polar acceptor strength (sa), polar donor strength of Hydrogens
    bonded to Carbon (sdc) or other non-Hydrogen atoms (sdx), number of lone
    pairs (num_lp), atomic number (z), and formal charge (q).

    Args:
    rdkit_molecule: RDKit molecule
    kallisto_molecule: kallisto molecule
    atoms_and_nbrs: list of lists of atoms and their bonded atoms
    charges: list of atomic partial charges
    d: parameter that ensures that the donor strength (sdx) of
    Hydrogen in H2O is equal to 1.000
    a: parameter that ensures that the acceptor strength (sa) of
    one lone pair on Oxygen in H2O is equal to 1.000
    t: bond reduction factor (fixed to ensure conditions above)

    Returns:
    Map of molecular polar properties (plus some extras)

    """
    # atomic numbers (e.g., C=6, H=1)
    ats = [atom.GetAtomicNum() for atom in rdkit_molecule.GetAtoms()]
    # atomic covalent coordination numbers
    cns = kallisto_molecule.get_cns(cntype="cov")
    # atomic static polarizabilities
    alps = getPolarizabilities(ats, cns, charges, 0)  # type: ignore

    mol_map = dict()
    for idx, atom in enumerate(rdkit_molecule.GetAtoms()):
        sa = 0.0
        sdx = 0.0
        sdc = 0.0
        nlps = 0
        z = ats[idx]
        q = atom.GetFormalCharge()
        eeq = charges[idx]
        hyb = atom.GetHybridization().name.lower()
        alp = alps[idx]

        # if Hydrogen -> evaluate donor strength
        if z == 1:
            sd = get_donor_atom_strength(idx, atoms_and_nbrs, charges, d, t)

            # distinguish between Carbon and other non-Hydrogen atoms
            nbr = atom.GetNeighbors()
            # Carbon case
            if nbr[0].GetAtomicNum() == 6:
                sdc = sd
            # other non-Hydrogen case
            else:
                sdx = sd
        else:
            nlps = get_lone_pairs(atom)
            if nlps > 0:
                sa = get_acceptor_atom_strength(idx, atoms_and_nbrs, charges, a, t)

        # create dict where key is atom index and values are properties
        atom_dict = {
            "z": z,
            "q": q,
            "eeq": round(eeq, ROUNDING_DIGITS),
            "alp": round(alp, ROUNDING_DIGITS),
            "hyb": hyb,
            "num_lp": nlps,
            "sdc": round(sdc, ROUNDING_DIGITS),
            "sdx": round(sdx, ROUNDING_DIGITS),
            "sa": round(sa, ROUNDING_DIGITS),
        }

        mol_map[idx] = atom_dict
    return mol_map


def get_lone_pairs(atom) -> int:
    """Get the number of lone pairs for an atom.

    The method is similar to that in
    https://github.com/rdkit/blob/master/Code/GraphMol/Aromaticity.cpp with
    the exception that it calculates the explicit valence via atom functions
    (not PeriodicTable). This is because the original method produces
    miscalculations for some systems (e.g., in molecules with SO2 groups).

    """
    # set up a periodic table
    pt = Chem.GetPeriodicTable()
    symbol = atom.GetSymbol()
    valence_electrons = PeriodicTable.GetNOuterElecs(pt, symbol)
    unavailable_electrons = atom.GetExplicitValence()
    charge = atom.GetFormalCharge()
    free_electrons = valence_electrons - unavailable_electrons - charge
    return int(free_electrons / 2)


def get_donor_atom_strength(
    atom_idx: int, atoms_and_nbrs: list, charges: list, d=63.7, t=0.274
) -> float:
    """Donor strength calculation - equation 10."""
    q, q_delta = get_atom_and_nbrs_charges(atom_idx, atoms_and_nbrs, charges, t)
    return d * (q + q_delta)


def get_acceptor_atom_strength(
    atom_idx: int, atoms_and_nbrs: list, charges: list, a=-4.4362, t=0.274
) -> float:
    """Acceptor strength calculation - equation 12 and 13."""
    q, q_delta = get_atom_and_nbrs_charges(atom_idx, atoms_and_nbrs, charges, t)
    return a * (q + q_delta)


def get_atom_neighbours(atom_idx: int, molecule_covalent_nbrs: list):
    """Get all neighbours for atom_idx.

    Extract all covalent bonding partner (alpha), all nearest neighbours
    (beta), and all nearest-nearest neighbours (gamma) for atom with index
    'atom_idx'.

    Args:
    atom_idx: index of atom to extract neighbours for
    molecule_covalent_list: list of covalent partners of each atom

    Returns:
    alpha: list of all covalent bonding atom indices of atom_idx
    beta: list of nearest neighbour atom indices of atom_idx
    gamma: list of nearest-nearest neighbour atom indices of atom_idx

    """
    # extract alpha neighbours
    alpha = molecule_covalent_nbrs[atom_idx]

    # extract beta neighbours
    beta = list()
    for _, a in enumerate(alpha):
        b = molecule_covalent_nbrs[a]
        diff = list(set([atom_idx]) ^ set(b))
        if len(diff) > 0:
            beta.extend(diff)

    # extract gamma neighbours
    gamma = list()
    for _, b in enumerate(beta):
        c = molecule_covalent_nbrs[b]
        inter = list(set(alpha).intersection(set(c)))
        diff = list(set(inter) ^ set(c))
        gamma.extend(diff)
    gamma = list(dict.fromkeys(gamma))
    return alpha, beta, gamma


def get_charges_from_atom_list(atom_idxs: list, charges: list) -> list:
    """Generic list-based function for charge retrieval."""
    return [charges[idx] for idx in atom_idxs]


def get_atom_and_nbrs_charges(
    atom_idx: int, atoms_and_nbrs: list, charges: list, t=0.274
):
    """Calculate charge and delta.

    Abstraction that includes q retrieval and calculation of Equation 9.
    The function can be used for both donor and acceptor strength calculations.

    """
    # get charges
    q = get_charges_from_atom_list([atom_idx], charges)[0]

    # get alpha, beta, and gamma neighbours and their charges
    alpha, beta, gamma = get_atom_neighbours(atom_idx, atoms_and_nbrs)
    q_alpha = get_charges_from_atom_list(alpha, charges)
    q_beta = get_charges_from_atom_list(beta, charges)
    q_gamma = get_charges_from_atom_list(gamma, charges)

    # equation 9
    q_delta = t * sum(q_alpha) + (t ** 2) * sum(q_beta) + (t ** 3) * sum(q_gamma)
    return q, q_delta


def calculate_delta_apolar(
    rdkit_molecule: Chem.rdchem.Mol,
    mol_map: dict,
    g0: float,
    gs: float,
    gr: float,
    gpi1: float,
    gpi2: float,
) -> float:
    """Calculate apolar free energy contribution.

    Calculate the apolar contribution to the free energy.
    Equation 7-8 - parameter names are adapted.

    Args:
    rdkit_molecule: RDKit molecule
    mol_map: molecular map of polar properties
    g0: zeroth order parameter
    gs: surface parameter
    gr: ring parameter
    gpi1: first pi parameter
    gpi2: second pi parameter

    Returns:
    Apolar free energy contribution (float)

    """
    # get atoms from rdkit molecule
    atoms = rdkit_molecule.GetAtoms()
    # atomic numbers (e.g., C=6, H=1)
    at = [atom.GetAtomicNum() for atom in atoms]
    # hybridization states
    hi = [atom.GetHybridization() for atom in atoms]
    # ring count
    ring = rdMolDescriptors.CalcNumRings(rdkit_molecule)
    # non-hydrogen count
    idxs = [idx for idx, nuclear_charge in enumerate(at) if nuclear_charge != 1]
    # atomic dynamic polarizabilities
    alp = [atom["alp"] for atom in mol_map.values()]
    # van der Waals radii in Angstrom
    vdws = getVanDerWaalsRadii(
        len(at),
        at,  # type: ignore
        alp,  # type: ignore
        vdwtype="rahm",
        scale=Bohr,
    )

    ns = 0
    nl = dict()
    for _, idx in enumerate(idxs):
        # non-hydrogen partner count
        nl[idx] = 0
        vdw = vdws[idx]
        atom = atoms[idx]
        nbrs = atom.GetNeighbors()
        for nbr in nbrs:
            if nbr.GetAtomicNum() != 1:
                nl[idx] += 1

        # sphere term ns
        bracket = 1 - (nl[idx]) / (hi[idx] + 1)
        ns += 4 * np.pi * np.power(vdw, 2) * bracket

    # zeroth order term
    dga = g0
    # surface term
    dga += gs * ns
    # ring term
    dga += gr * ring

    # identify pi orbitals in hybridization states
    # SP  : 2 p-orbitals
    # SP2 : 1 p-orbital
    sp, sp2 = 0, 0
    for _, v in enumerate(hi):
        if v == rdchem.HybridizationType.SP:
            sp += 2
        elif v == rdchem.HybridizationType.SP2:
            sp2 += 1

    # pi-orbital terms
    dga += sp * gpi1
    dga += sp2 * gpi2
    return float(dga)


def calculate_delta_polar(
    mol_map: dict,
    atoms_and_nbrs: list,
    gd: float,
    ga: float,
    expd: float,
    expa: float,
) -> float:
    """Calculate polar contribution to free energy.

    Calculate the polar contribution to the free energy.
    Equation 14 - parameter names adapted.

    Args:
    mol_map: molecular map of polar properties as obtained by
    the `calculate_polar_strength_map()` function
    atoms_and_nbrs: list of lists of atoms and their bonded atoms
    gd: hydrogen bond donor parameter
    ga: hydrogen bond acceptor parameter
    expd: exponent for hydrogen bond donor
    expa: exponent for hydrogen bond acceptor

    Returns:
    Polar free energy contribution (float)

    """
    don = 0.0
    acc = 0.0

    for _, idx in enumerate(mol_map):
        atom = mol_map[idx]
        nh = 0
        sdi = 0.0
        nlps = 0
        sak = 0.0

        # if atom is Hydrogen -> skip
        # get contribution from heavy partner atom
        if atom["z"] == 1:
            continue

        # donor contribution
        nbrs_idxs = atoms_and_nbrs[idx]
        for _, nbr_idx in enumerate(nbrs_idxs):
            nbr = mol_map[nbr_idx]

            # if neighbor is Hydrogen
            # get strength and add to total
            if nbr["z"] == 1:
                sd = nbr["sdx"] + nbr["sdc"]
                sdi += sd
                nh += 1

        # mean donor strength
        if nh > 0:
            sdi = sdi / nh

        # acceptor contribution
        nlps = atom["num_lp"]
        sak = atom["sa"]

        don += sdi * (nh ** expd)
        acc += sak * (nlps ** expa)
    # equation 14
    return float(gd * don + ga * acc)


def any_hydrogen_neighbors(rdkit_atom: Chem.rdchem.Atom):
    """Returns True is Hydrogen is a neighbor else False.

    Written because `rdkit.Chem.rdchem.Atom.GetTotalHs()` does not work on
    embedded molecules.

    Args:
    rdkit_atom: rdkit.Chem.rdchem.Atom object

    Returns:
    Boolean

    """
    for nbr in rdkit_atom.GetNeighbors():
        if nbr.GetAtomicNum() == 1:
            return True
    return False


def interaction_strength(idx: int, mol_map: dict, acceptor_exp: float) -> float:
    """Calculate interaction strength for atom with index `idx`."""
    acceptor_strength = mol_map[idx]["sa"]
    num_lp = mol_map[idx]["num_lp"]
    if num_lp != 0:
        return acceptor_strength * (num_lp ** acceptor_exp)
    return 0.0


def calculate_delta_interaction(
    rdkit_molecule: Chem.rdchem.Mol,
    mol_map: dict,
    atoms_and_nbrs: list,
    gi: float,
    expa: float,
    f: float,
):
    """Calculate interaction contribution to free energy.

    Calculate the interation contribution to the free energy.
    Equation 15 - parameter names adapted.

    Args:
    rdkit_molecule: RDKit molecule
    mol_map: molecular map of polar properties as obtained by
    atoms_and_nbrs: list of lists of atoms and their bonded atoms
    gd: interaction parameter
    f: correction parameter

    Returns:
    Interaction free energy contribution (float)

    """
    # extract neighbors
    alpha, beta, gamma = get_all_neighbors(rdkit_molecule, atoms_and_nbrs)

    dgi = 0.0
    atoms = rdkit_molecule.GetAtoms()
    for idx, _ in enumerate(atoms):
        # strength of origin
        s_origin = interaction_strength(idx, mol_map, expa)

        # if any Hydrogen is available continue
        if any_hydrogen_neighbors(atoms[idx]):
            continue

        # alpha neighbor contributions
        for _, va in enumerate(alpha[idx]):
            if any_hydrogen_neighbors(atoms[va]):
                continue

            # strength of partner
            s_partner = interaction_strength(va, mol_map, expa)
            dgi += s_origin * s_partner

        # beta neighbor contribution
        for _, vb in enumerate(beta[idx]):
            if any_hydrogen_neighbors(atoms[vb]):
                continue

            # strength of partner
            s_partner = interaction_strength(vb, mol_map, expa)
            dgi += s_origin * s_partner

        # gamma neighbor contribution
        for _, vc in enumerate(gamma[idx]):
            if any_hydrogen_neighbors(atoms[vc]):
                continue

            # strength of partner
            s_partner = interaction_strength(vc, mol_map, expa)
            dgi += f * s_origin * s_partner
    return gi * dgi



"""RDKit functions to convert Jazzy data into rendering."""
# src/jazzy/visualisation.py
from typing import Tuple

import rdkit
from rdkit import Chem
from rdkit.Chem.Draw import rdMolDraw2D

from jazzy.config import ANNOTATION_FONT_SCALE
from jazzy.config import ROUNDING_DIGITS


def _zero_positive_value_check(v: float):
    """Assert that the input is zero or positive.

    Raises:
    ValueError: An error if the input is negative

    """
    if v < 0:
        raise ValueError("The value should be zero or positive (got {})".format(v))


def _fix_explicit_hydrogens(rwmol: Chem.rdchem.RWMol, idx: int):
    """Increase number of explicit hydrogens for atom with index `idx`.

    Args:
    rwmol: An RDKit RWmolecule (rdkit.Chem.rdchem.RWMol)
    idx: index of atom (int)

    """
    oa = rwmol.GetAtomWithIdx(idx)
    if oa.GetAtomicNum() > 1:
        oa.SetNumExplicitHs(oa.GetNumExplicitHs() + 1)


def _remove_excluded_hydrogens(rwmol: Chem.rdchem.RWMol, excluded_hydrogens: list):
    """Convert list of hydrogen atoms into explicit hydrogens.

    Convert implicit hydrogen atoms into explicit hydrogen on the atoms where
    they are bonded. E.g., H-N-H is converted to NH2.

    Args:
    rwmol: An RDKit RWmolecule (rdkit.Chem.rdchem.RWMol)
    excluded_hydrogens: list of implicit Hydrogens to convert into explicit ones.

    Returns:
    rdkit_molecule: An RDKit molecule (rdkit.Chem.rdchem.Mol)

    """
    if not excluded_hydrogens:
        return rwmol

    ai_to_remove = list()
    for bond_idx in reversed(range(rwmol.GetNumBonds())):
        b = rwmol.GetBondWithIdx(bond_idx)
        bidx = b.GetBeginAtomIdx()
        remove_bidx = bidx in excluded_hydrogens
        eidx = b.GetEndAtomIdx()
        remove_eidx = eidx in excluded_hydrogens
        if remove_bidx or remove_eidx:
            if remove_bidx:
                ai_to_remove.append(bidx)
                _fix_explicit_hydrogens(rwmol, eidx)
            if remove_eidx:
                ai_to_remove.append(eidx)
                _fix_explicit_hydrogens(rwmol, bidx)
            rwmol.RemoveBond(bidx, eidx)
    for atom_idx in sorted(ai_to_remove, reverse=True):
        rwmol.RemoveAtom(atom_idx)
    rdkit.Chem.SanitizeMol(rwmol)
    return rwmol.GetMol()


def _remove_strong_acceptor_hydrogens(rwmol: Chem.rdchem.RWMol, hs_to_remove: list):
    """Remove Hydrogens that are bonded to strong acceptors.

    Goes through a list of hydrogens and checks wheather any of their neighbors
    has been annotated with acceptor strength. If so, it removes the hydrogen
    from the removal list to avoid it to collapse with the acceptor atom.

    Args:
    rwmol: An RDKit RWmolecule (rdkit.Chem.rdchem.RWMol)
    hs_to_remove: list of implicit Hydrogens to convert into explicit ones.

    Returns:
    Updated list for hydrogen atoms to be removed.

    """
    updated_hs_to_remove = list()
    for idx in hs_to_remove:
        hs = rwmol.GetAtomWithIdx(idx)
        nbrs = hs.GetNeighbors()
        for nbr in nbrs:
            if not nbr.HasProp("sa"):
                updated_hs_to_remove.append(idx)
                break
    return updated_hs_to_remove


def _create_color_scale(idx2strength: dict, mode: str):
    """Create RGB mapping.

    Normalises values in a dictionary and creates an RGB red or blue scale
    mapping for them. `abs(round(0.9-alpha, ROUNDING_DIGITS))` is a bit hacky
    as it avoids the creation of whites by sacrificing a bit of alpha.
    Also RGB for acceptor has been tweaked a bit to avoid generating scales
    that might hide atoms rendered with blue fonts (e.g., Nitrogen atoms).

    Args:
    idx2strength: mapping between indices and strengths (dict)
    mode: `acceptor` or `donor` (str)

    Returns:
    Mapping between indices and RGB colors (dict).

    """
    valid_modes = ["donor", "acceptor"]
    if mode not in valid_modes:
        raise ValueError("Select valid mode. Either 'donor' or 'acceptor'.")

    strengths = idx2strength.values()
    if len(strengths) == 0:
        return idx2strength
    amin, amax = min(strengths), max(strengths)
    idx2rgb = dict()
    for idx, v in idx2strength.items():
        try:
            alpha = round((v - amin) / (amax - amin), ROUNDING_DIGITS)
        except ZeroDivisionError:
            alpha = round(v, ROUNDING_DIGITS)
        invalpha = abs(round(0.9 - alpha, ROUNDING_DIGITS))
        if mode == "donor":
            rgb = (1.0, invalpha, invalpha)
        else:
            # acceptor
            rgb = (invalpha, 0.7, 1.0)
        idx2rgb[idx] = rgb
    return idx2rgb


def _draw_molecule(
    rwmol: Chem.rdchem.RWMol,
    fig_size: Tuple[int, int],
    atoms_to_highlight: list,
    idx2rgb: dict,
):
    """Draw an RDKit molecule.

    Args:
    rwmol: An RDKit RWmolecule (rdkit.Chem.rdchem.RWMol)

    Returns:
    Text of an SVG image.

    """
    d2d = rdMolDraw2D.MolDraw2DSVG(fig_size[0], fig_size[1])
    d2d.drawOptions().annotationFontScale = ANNOTATION_FONT_SCALE
    d2d.drawOptions().padding = 0.15
    d2d.DrawMolecule(
        rwmol,
        highlightAtoms=atoms_to_highlight,
        highlightAtomColors=idx2rgb,
        highlightBonds=None,
    )
    d2d.FinishDrawing()
    return d2d.GetDrawingText()


def _set_acceptor_props(
    atom: rdkit.Chem.rdchem.Atom, sa: float, sa_threshold: float, ignore_sa: bool
) -> bool:
    """Setting two props to acceptor to avoid mistakes in the highlighting function.

    Args:
    atom: An RDKit atom (rdkit.Chem.rdchem.Atom)
    sa: acceptor strength (float)
    sa_threshold: threshold to show acceptor contribution (float)
    ignore_sa: ignore acceptor contributions (bool)

    Returns:
    If properties have been set to atom (bool).

    """
    condition = False
    if sa != 0 and sa > sa_threshold and not ignore_sa:
        atom.SetProp("atomNote", str(sa))
        atom.SetProp("sa", str(sa))
        condition = True
    return condition


def _set_donor_props(
    atom: rdkit.Chem.rdchem.Atom, sd: float, sd_threshold: float, ignore_sd: bool
) -> bool:
    """Setting two props to donor to avoid mistakes in the highlighting function.

    Args:
    atom: An RDKit atom (rdkit.Chem.rdchem.Atom)
    sd: donor strength (float)
    sd_threshold: threshold to show donor contribution (float)
    ignore_sd: ignore donor contributions (bool)

    Returns:
    If properties have been set to atom (bool).

    """
    condition = False
    if sd != 0 and sd > sd_threshold and not ignore_sd:
        atom.SetProp("atomNote", str(sd))
        atom.SetProp("sd", str(sd))
        condition = True
    return condition


def _exclude_hydrogens(
    rwmol: Chem.rdchem.RWMol,
    atomic_map: dict,
    sa_threshold: float,
    sdc_threshold: float,
    sdx_threshold: float,
    ignore_sa: bool,
    ignore_sdc: bool,
    ignore_sdx: bool,
    rounding_digits=ROUNDING_DIGITS,
):
    """Create `excluded_hydrogens` list from RDKit molecule.

    Args:
    rwmol: An RDKit RWmolecule (rdkit.Chem.rdchem.RWMol)
    atomic_map: molecular map of polar properties (dict)
    sa_threshold: threshold to show acceptor contribution (float)
    sdc_threshold: threshold to show donor contribution on Carbon (float)
    sdx_threshold: threshold to show donor contribution on non-Carbon (float)
    ignore_sa: ignore acceptor contributions (bool)
    ignore_sdc: ignore donor contributions on Carbon (bool)
    ignore_sdx: ignore donor contributions on non-Carbon (bool)

    Returns:
    `excluded_hydrogens` list.

    """
    excluded_hydrogens = list()
    for idx, atom in enumerate(rwmol.GetAtoms()):
        atom_props = atomic_map[idx]

        # not involved in hydrogen bonding
        sa = round(atom_props["sa"], rounding_digits)
        sdc = round(atom_props["sdc"], rounding_digits)
        sdx = round(atom_props["sdx"], rounding_digits)
        if sa == 0.0 and sdc == 0.0 and sdx == 0.0:
            continue

        # acceptor logic
        if _set_acceptor_props(atom, sa, sa_threshold, ignore_sa):
            continue

        # donor logic for carbons
        if _set_donor_props(atom, sdc, sdc_threshold, ignore_sdc):
            continue

        # donor logic for non-carbons
        if _set_donor_props(atom, sdx, sdx_threshold, ignore_sdx):
            continue

        # hydrogen does not fall into above conditions
        # then not strong enough and mark for removal
        if atom_props["z"] == 1:
            excluded_hydrogens.append(idx)

    # remove strong acceptor hydrogens
    if not ignore_sa:
        excluded_hydrogens = _remove_strong_acceptor_hydrogens(
            rwmol, excluded_hydrogens
        )

    return excluded_hydrogens


def _get_highlighted_atoms_and_strength_colors(
    rdkit_molecule: Chem.rdchem.Mol,
    highlight_atoms: bool,
):
    """Get donor and acceptor strengths and create RGB colors from them.

    Args:
    rdkit_molecule: An RDKit RWmolecule (rdkit.Chem.rdchem.RWMol)
    highlight_atoms: show donors in red and acceptors in blue (bool)

    Returns:
    `atoms_to_highlight`, `idx2sa`, `idx2sd`, and `idx2rgb` lists.

    """
    idx2sd = dict()
    idx2sa = dict()
    idx2rgb = dict()
    atoms_to_highlight = list()
    if highlight_atoms:
        atoms = [x for x in rdkit_molecule.GetAtoms()]
        for a in atoms:
            if a.HasProp("sd"):
                idx = a.GetIdx()
                atoms_to_highlight.append(idx)
                idx2sd[idx] = float(a.GetProp("sd"))
            if a.HasProp("sa"):
                idx = a.GetIdx()
                atoms_to_highlight.append(idx)
                idx2sa[idx] = float(a.GetProp("sa"))

        # convert strengths into relative RGB scale and combine the scales
        idx2reds = _create_color_scale(idx2sd, mode="donor")
        idx2blues = _create_color_scale(idx2sa, mode="acceptor")
        idx2rgb = {**idx2reds, **idx2blues}  # type: ignore
    return atoms_to_highlight, idx2rgb


def depict_strengths(
    rdkit_molecule: Chem.rdchem.Mol,
    atomic_map: dict,
    fig_size=(500, 500),
    flatten_molecule=False,
    highlight_atoms=False,
    ignore_sdc=False,
    ignore_sdx=False,
    ignore_sa=False,
    sdc_threshold=0.0,
    sdx_threshold=0.0,
    sa_threshold=0.0,
    rounding_digits=ROUNDING_DIGITS,
):
    """Create an SVG image text from an RDKit molecule and its atomic map.

    The default configuration simply produces a depiction of the input molecule
    and its strengths. `highlight_atoms` highlights atoms in red (donors) and
    blue (acceptors). `flatten_molecule` produces a 2-dimensional depiction of
    the molecule. Any `threshold` parameter allows to set a numeric threshold
    under which strengths are not included in the output depiction.

    Args:
    rdkit_molecule: An RDKit molecule (rdkit.Chem.rdchem.Mol)
    atomic_map: molecular map of polar properties (dict)
    fig_size: size of the depiction in pixels (tuple)
    flatten_molecule: boolean to create 2D depiction (bool)
    highlight_atoms: show donors in red and acceptors in blue (bool)
    ignore_sdc: ignore donor contributions on Carbon (bool)
    ignore_sdx: ignore donor contributions on non-Carbon (bool)
    ignore_sa: ignore acceptor contributions (bool)
    sdc_threshold: threshold to show donor contribution on Carbon (float)
    sdx_threshold: threshold to show donor contribution on non-Carbon (float)
    sa_threshold: threshold to show acceptor contribution (float)

    Returns:
    SVG depiction of given RDKit molecule and its atomic strength map.

    """
    # assert that thresholds are zero or positive
    for v in [sdc_threshold, sdx_threshold, sa_threshold]:
        _zero_positive_value_check(v)

    # copy RDKit molecule otherwise the input would be affected by processing
    rwmol = rdkit.Chem.RWMol(rdkit_molecule)

    # flatten the molecule if required
    # Note: never create SMILES then again RDKit 2D molecule since the mapping
    # could be shuffled! Only recompute 2D coordinates on the 3D molecule.
    if flatten_molecule:
        rdkit.Chem.rdDepictor.Compute2DCoords(rwmol)

    # create list of excluded hydrogens
    excluded_hydrogens = _exclude_hydrogens(
        rwmol,
        atomic_map,
        sa_threshold,
        sdc_threshold,
        sdx_threshold,
        ignore_sa,
        ignore_sdc,
        ignore_sdx,
        rounding_digits,
    )

    # remove excluded hydrogen atoms
    rwmol = _remove_excluded_hydrogens(rwmol, excluded_hydrogens)

    # find atoms with properties appended as all indices have changed now
    (
        atoms_to_highlight,
        idx2rgb,
    ) = _get_highlighted_atoms_and_strength_colors(rwmol, highlight_atoms)

    # draw the molecule and return render
    img_txt = _draw_molecule(
        rwmol=rwmol,
        fig_size=fig_size,
        atoms_to_highlight=atoms_to_highlight,
        idx2rgb=idx2rgb,
    )
    return img_txt



"""Anything else."""
# src/jazzy/utils.py


class JazzyError(Exception):
    """Raised when Jazzy cannot calculate results."""

    pass


class KallistoError(Exception):
    """Raised when something goes wrong inside kallisto."""

    pass



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['jazzy']

package_data = \
{'': ['*']}

install_requires = \
['click<8.0.0', 'kallisto>=1.0.7,<2.0.0', 'rdkit-pypi<=2021.9.4']

entry_points = \
{'console_scripts': ['jazzy = jazzy.__main__:cli']}

setup_kwargs = {
    'name': 'jazzy',
    'version': '0.0.6',
    'description': 'Jazzy',
    'long_description': ".. image:: https://raw.githubusercontent.com/AstraZeneca/jazzy/master/docs/_static/jazzy.png\n  :width: 400\n  :alt: Jazzy\n\n|PyPI| |Status| |Python Version| |License|\n\n|Read the Docs| |Tests| |Codecov|\n\n|pre-commit| |Black|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/jazzy.svg\n   :target: https://pypi.org/project/jazzy/\n   :alt: PyPI\n.. |Status| image:: https://img.shields.io/pypi/status/jazzy.svg\n   :target: https://pypi.org/project/jazzy/\n   :alt: Status\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/jazzy\n   :target: https://pypi.org/project/jazzy\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/pypi/l/jazzy\n   :target: https://opensource.org/licenses/Apache-2.0\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/jazzy/latest.svg?label=Read%20the%20Docs\n   :target: https://jazzy.readthedocs.io/\n   :alt: Read the documentation at https://jazzy.readthedocs.io/\n.. |Tests| image:: https://github.com/AstraZeneca/jazzy/workflows/Tests/badge.svg\n   :target: https://github.com/AstraZeneca/jazzy/actions?workflow=Tests\n   :alt: Tests\n.. |Codecov| image:: https://codecov.io/gh/AstraZeneca/jazzy/branch/master/graph/badge.svg?token=4HCWYH61S5\n   :target: https://codecov.io/gh/AstraZeneca/jazzy\n   :alt: Codecov\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n\nFull Author List\n----------------\n* Eike Caldeweyher\n* Gian Marco Ghiandoni\n\n\nIntroduction\n------------\n*Jazzy* is an efficient computational tool for the calculation of hydration free energies and hydrogen bond acceptor and donor strengths.\n\n\nFeatures\n--------\n\n* Hydration free energies\n* Hydrogen bond strengths\n* Visualisation functions (SVG)\n* Application programming interface\n* Command-line interface\n\n\nRequirements\n------------\n\n.. code:: console\n\n   click==7.1.2            # Composable command line interface toolkit\n   kallisto==1.0.7         # Atomic and molecular featurizer\n   numpy==1.21.1           # NumPy array computing package\n   rdkit-pypi<=2021.9.4    # Chemoinformatics toolkit\n\n\nInstallation\n------------\n\nYou can install *Jazzy* via pip_ from PyPI_:\n\n.. code:: console\n\n   $ pip install jazzy\n\n\nInstallation from Source\n------------------------\n\nRequirements to install *Jazzy* from sources:\n\n- `poetry`_\n- `pyenv`_ or `conda`_\n- python>=3.6\n\nFirst check that ``poetry`` is running correctly (v1.0.10 at the time of writing)\n\n.. code:: console\n\n   $ poetry --version\n   Poetry version 1.0.10\n\nCreate a virtual environment (via ``pyenv`` or ``conda``) and activate it. Afterwards, clone the *Jazzy* project from GitHub and install it using ``poetry``\n\n.. code:: console\n\n   $ git clone git@github.com:AstraZeneca/jazzy.git\n   $ cd jazzy\n   $ poetry install\n\n\nUsage and Cookbook\n------------------\n\nPlease see the `Usage <Usage_>`_ and `Cookbook <Cookbook_>`_ sections for details.\n\n\nContributing\n------------\n\nJazzy is an open project in every shape and form, thus feedback on how to improve its documentation or functionalities is always welcome.\nTo learn more, see the `Contributor Guide`_.\n\n\nLicense\n-------\n\nDistributed under the terms of the `Apache 2.0 license`_,\n*Jazzy* is free and open source software.\n\n\nIssues\n------\n\nIf you encounter any problems,\nplease `file an issue`_ along with a detailed description.\n\n\nCredits\n-------\n\nThis project was generated from `@cjolowicz`_'s `Hypermodern Python Cookiecutter`_ template.\n\n.. _@cjolowicz: https://github.com/cjolowicz\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _Apache 2.0 license: https://opensource.org/licenses/Apache-2.0\n.. _poetry: https://python-poetry.org/docs/#installation\n.. _pyenv: https://github.com/pyenv/pyenv#installation\n.. _conda: https://docs.conda.io/projects/conda/en/latest/user-guide/install/index.html\n.. _PyPI: https://pypi.org/\n.. _Hypermodern Python Cookiecutter: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n.. _file an issue: https://github.com/AstraZeneca/jazzy/issues\n.. _pip: https://pip.pypa.io/\n.. github-only\n.. _Contributor Guide: contributing.rst\n.. _Cookbook: https://jazzy.readthedocs.io/en/latest/cookbook.html\n.. _Usage: https://jazzy.readthedocs.io/en/latest/usage.html\n",
    'author': 'Gian Marco Ghiandoni',
    'author_email': 'ghiandoni.g@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/AstraZeneca/jazzy',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.0,<4.0.0',
}


setup(**setup_kwargs)
